from .preprocess import *
from .data_loading import *
from .basicNetwork import basicNetwork
from .string_manipulate import *
from .opt_gaussian import *
from .data_stats import *
from .IO import *
from .dependency import *
from .tools.DManager import *



#!/usr/bin/env python

import numpy as np
import torch

def pretty_np_array(m, front_tab='', verticalize=False):
	if type(m) == type(torch.tensor([])):
		m = m.cpu().detach().numpy()

	if verticalize:
		if len(m.shape) == 1:
			m = np.atleast_2d(m).T

	out_str = front_tab + str(m).replace('\n ','\n' + front_tab).replace('[[','[').replace(']]',']') + '\n'
	#out_str = str(out_str).replace('. ',' ')
	out_str = str(out_str).replace('.]',']')
	return out_str

def block_two_string_concatenate(str1, str2, spacing='\t'):
	L1 = str1.split('\n')
	L2 = str2.strip().split('\n')

	if len(L1) > len(L2):
		Δ = len(L1) - len(L2)
		for ι in range(Δ):
			L2.append('\n')

	max_width = len(max(L1, key=len))
	outS = ''
	for l1, l2 in zip(L1,L2):
		outS += ('%-' + str(max_width) + 's' + spacing + l2 + '\n') % l1

	return outS


def block_string_concatenate(strList, spacing='\t'):
	all_concat = ''
	for single_block in strList:

		if all_concat == '':
			all_concat = single_block
		else:
			all_concat = block_two_string_concatenate(all_concat, single_block, spacing=spacing)

	return all_concat


if __name__ == "__main__":
	str1 = 'Hello world\nA interesting string'
	str2 = 'Second Block\nto the right\nTooLong'
	outS = block_two_string_concatenate(str1, str2)
	print(outS)

	str1 = 'Hello world\nA interesting string'
	str2 = 'Second Block, to the right, TooLong'
	str3 = 'Third Block, \n make it shorter'
	all_3 = [str1, str2, str3]
	outS = block_string_concatenate(all_3)
	print(outS)






import numpy as np; np.random.seed(0)
from wplotlib import lines		#pip install wplotlib
from wplotlib import heatMap
from wplotlib import histograms
import wuml 
from wuml.data_loading import load_csv

import io
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns; sns.set_theme()
import random


def get_feature_histograms(X, path=None, title=''):
	header = './results/DatStats/'
	wuml.ensure_path_exists('./results')
	wuml.ensure_path_exists(header)

	H = histograms()
	H.histogram(X, num_bins=10, title=title, fontsize=12, facecolor='blue', α=0.5, path=path, subplot=None)

def identify_missing_data_per_feature(df):
	df = df.get_data_as('DataFrame')

	X = df.values
	n = X.shape[0]
	d = X.shape[1]

	mdp = missing_data_counter_per_feature = []
	for column in df:
		x = df[column].values
		missing_percentage = np.sum(np.isnan(x))/n
		mdp.append(missing_percentage)

	return mdp

def missing_data_stats(df): 
	df = df.get_data_as('DataFrame')

	header = './results/DatStats/'
	wuml.ensure_path_exists('./results')
	wuml.ensure_path_exists(header)

	mdp = identify_missing_data_per_feature(df)
	mdp = np.array(mdp)
	textstr = ''
	x = np.arange(1, len(mdp)+1)

	lp = lines()
	lp.plot_line(x, mdp, 'Missing Percentage', 'Feature ID', 'Percentage Missing', 
				imgText=textstr, outpath= header + 'feature_missing_percentage.png')


	X2 = np.isnan(X).astype(int)
	hMap = heatMap()
	hMap.draw_HeatMap(X2, title='Missing Data Heat Map', 
							xlabel='Feature ID', ylabel='Sample ID',
							path= header + 'missing_data_heatMap.png')

	buffer = io.StringIO()
	df.info(buf=buffer, verbose=True)
	s = buffer.getvalue()
	wuml.write_to(s, header + 'feature_stats.txt')







from sklearn import preprocessing
import wuml
import pandas as pd
import numpy as np
import torch

class load_csv:
	def __init__(self, xpath=None, ypath=None, dataFrame=None, X_npArray=None, row_id_with_label=None , 
					sample_id_included=False, torchDataType=torch.FloatTensor, batch_size=20):
		'''
			row_id_with_label :  None = no labels, 0 = top row is the label
			dataFrame: if dataFrame is set, it ignores the path and use the dataFrame directly as the data itself
		'''
		if dataFrame is not None:
			self.df = dataFrame
		elif X_npArray is not None:
			self.df = pd.DataFrame(X_npArray)
		else:
			self.df = pd.read_csv (xpath, header=row_id_with_label)

		if ypath is not None: 
			self.Y = np.loadtxt(ypath, delimiter=',', dtype=np.float32)			

		self.batch_size = batch_size
		self.shape = self.df.shape
		self.torchDataType = torchDataType				
		self.torchloader = None

	def get_data_as(self, data_type): #'DataFrame', 'read_csv', 'Tensor'
		if data_type == 'read_csv': return self
		if data_type == 'DataFrame': return self.df
		if data_type == 'ndarray': 
			#self.df.values[subset]
			return self.df.values
		if data_type == 'DataLoader' and self.torchloader is None: 
			self.DM = wuml.DManager(self.df.values, self.Y)
			self.torchloader = DataLoader(dataset=self.DM, batch_size=self.batch_size, shuffle=True, pin_memory=True, num_workers=1)
			return self.torchloader

	def __getitem__(self, item):
		import pdb; pdb.set_trace()
		print('kkk')
		return self.df.values[item]

	def __str__(self): 
		print(self.df)
 



#!/usr/bin/env python

import torch
import numpy as np
import sys
from sklearn import linear_model
from torch import nn
from torch.autograd import Variable
import torch.nn.functional as F
import collections

class flexable_Model(torch.nn.Module):
	def __init__(self, dataDim, networkStructure, learning_rate=0.001):
		super(flexable_Model, self).__init__()

		inDim = dataDim
		for l, Layer_info in enumerate(networkStructure):
			layer_width, activation_function = Layer_info
			φ = activation_function
			outDim = layer_width

			lr = 'self.l' + str(l) + ' = torch.nn.Linear(' + str(inDim) + ', ' + str(outDim) + ' , bias=True)'
			exec(lr)
			exec('self.l' + str(l) + '.activation = "' + φ + '"')		#softmax, relu, tanh, sigmoid, none
			inDim = outDim

	def forward(self, x, y,ind):
		self.y0 = x
		for m, layer in enumerate(self.children(),0):
			var = 'self.y' + str(m+1)

			if layer.activation == 'none':
				cmd = 'self.yout = ' + var + ' = self.l' + str(m) + '(y' + str(m) + ')'
			else:
				cmd = 'self.yout = ' + var + ' = F.' + layer.activation + '(self.l' + str(m) + '(y' + str(m) + '))'
			exec(cmd)
		return self.yout

class basicNetwork:
	def __init__(self, costFunction, X, 
						Y=None, networkStructure=[(3,'relu'),(3,'relu'),(3,'none')], 
						on_new_epoch_call_back = None, max_epoch=1000, 	Torch_dataType=torch.FloatTensor, 
						learning_rate=0.001):
		'''
			possible activation functions: softmax, relu, tanh, sigmoid, none
		'''
		#	X should be in wuml format
		self.trainLoader = X.get_data_as('DataLoader')

		self.lr = learning_rate
		self.max_epoch = max_epoch
		self.Torch_dataType = Torch_dataType
		self.costFunction = costFunction
		self.NetStructure = networkStructure
		self.on_new_epoch_call_back = on_new_epoch_call_back #set this as a callback at each function
		self.model = flexable_Model(X.shape[1], networkStructure, learning_rate=0.001)

		self.info()

	def info(self):
		print('Network Info:')
		print('\tLearning rate: %.3f'%self.lr)
		print('\tNetwork Structure')
		for i in self.model.children():
			try:
				print('\t\t%s , %s'%(i,i.activation))
			except:
				print('\t\t%s '%(i))



	def on_new_epoch_call_back(self, loss_avg, num_of_epoch, lr):
		## Get Train, Test Accuracy, get loss, epoch, lr
		#db = self.db
		#db['debug'].output_current_network_state(loss_avg, num_of_epoch, lr, db['bigS'])
		#if loss_avg < 0.00001: return True	# early exit
		return False

	def train(self):
		model = self.model
	
		optimizer = torch.optim.Adam(model.parameters(), lr=self.lr)	
		scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( optimizer, factor=0.5, min_lr=1e-10, patience=50, verbose=False)
	
		for epoch in range(self.max_epoch):
	
			loss_list = []	
			for (i, data) in enumerate(self.trainLoader):
				[x, y, ind] = data
			
				x = Variable(x.type(self.Torch_dataType), requires_grad=False)
				y = Variable(y.type(self.Torch_dataType), requires_grad=False)
				optimizer.zero_grad()
				
				ŷ = model(x, y, ind)
				loss = self.costFunction(x, y, ŷ, ind)
				
				loss.backward()
				optimizer.step()
	
				loss_list.append(loss.item())

			loss_avg = np.array(loss_list).mean()
			scheduler.step(loss_avg)
			print(loss_avg)

		import pdb; pdb.set_trace()
			#if self.on_new_epoch_call_back is not None:
			#	early_exit = model.on_new_epoch(loss_avg, (epoch+1), scheduler._last_lr[0])
			#	if early_exit: break



